#ifndef _AFFIXMGR_HXX_
#define _AFFIXMGR_HXX_

#include "atypes.hxx"
#include "baseaffix.hxx"
#include "hashmgr.hxx"

class AffixMgr
{

  AffEntry *          pStart[SETSIZE];
  AffEntry *          sStart[SETSIZE];
  HashMgr *           pHMgr;
  char *              trystring;
  char *              encoding;
  char                compound;
  char                compound2;
  char                compound3;
  char *              compoundword;
  char                forbiddenword;
  int                 cpdmin;
  int                 numrep;
  replentry *         reptable;
  int                 cpdwordmax;
  int                 cpdmaxsyllable;
  char *              cpdvowels;
  replentry *         accent;
  char *              cpdsyllablenum;
  const char *        pfxappnd;
  const char *        sfxappnd;
  char	              sfxflag;
  int                 checknum;
  char *              wordchars;
  
public:
 
  AffixMgr(const char * affpath, HashMgr * ptr);
  ~AffixMgr();
  struct hentry *     affix_check(const char * word, int len);
  struct hentry *     prefix_check(const char * word, int len, int prefix_in_cpms);
  struct hentry *     suffix_check(const char * word, int len, int sfxopts, AffEntry* ppfx);
  int                 get_syllable (const char * word, int wlen);
  int                 repl_check(const char * word, int len);
  int                 candidate_check(const char * word, int len);
  struct hentry *     compound_check(const char * word, int len, char compound_flag, 
                              int wordnum, int numsyllable, int maxwordnum, int hu_mov_rule);
  struct hentry *     lookup(const char * word);
  int                 get_numrep();
  struct replentry *  get_reptable();
  struct replentry *  get_accent();
  char *              get_encoding();
  char *              get_try_string();
  char *              get_wordchars();
  char                get_compound();
  char                get_forbiddenword();
  char *              get_compoundword();
  int                 get_checknum();

private:
  int  parse_file(const char * affpath);
  int  parse_try(char * line);
  int  parse_set(char * line);
  int  parse_cpdflag(char * line);
  int  parse_cpdfirst(char * line);
  int  parse_cpdlast(char * line);
  int  parse_cpdword(char * line);
  int  parse_cpdforbid(char * line);
  int  parse_cpdmin(char * line);
  int  parse_cpdsyllable(char * line);
  int  parse_accent(char * line);
  int  parse_syllablenum(char * line);
  int  parse_reptable(char * line, FILE * af);
  int  parse_affix(char * line, const char at, FILE * af);
  int  parse_wordchars(char * line);

  int  rule119(const char * word, int len, 
      int * compound2stat, struct hentry * rv);

  void encodeit(struct affentry * ptr, char * cs);
  int build_pfxlist(AffEntry* pfxptr);
  int build_sfxlist(AffEntry* sfxptr);
  int process_pfx_order();
  int process_sfx_order();
};

#endif

